# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from .monte_carlo import MonteCarlo
from .predictor import Predictor
from .prediction import Prediction, UnweightedSamplesPrediction, PredictionResults
from .toe_prediction_profile import ToEPredictionProfile
from .unscented_transform import UnscentedTransformPredictor

# For naming consistancy
# Unfortunately, prog_algs was released with inconsistent naming (UnscentedTransformPredictor vs MonteCarlo).
# For naming consistency and to avoid confusion, we created aliases for the two classes.
# They can be called by the name of the method (e.g., UnscentedTranform) or with 'predictor' at the end (e.g., UnscentedTransformPredictor).
UnscentedTransform = UnscentedTransformPredictor
MonteCarloPredictor = MonteCarlo

__all__ = ['predictor', 'monte_carlo', 'unscented_transform', 'MonteCarlo', 'Predictor', 'Prediction', 'UnweightedSamplesPrediction', 'ToEPredictionProfile', 'UnscentedTransformPredictor', 'UnscentedTransform', 'MonteCarloPredictor']



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

"""
This file includes functions for calculating metrics given a Time of Event (ToE) profile (i.e., ToE's calculated at different times of prediction resulting from running prognostics multiple times, e.g., on playback data). The metrics calculated here are specific to multiple ToE estimates (e.g. alpha-lambda metric)
"""
from numpy import sign
from collections import defaultdict
from typing import Callable, Dict

from ..predictors import ToEPredictionProfile

def alpha_lambda(toe_profile : ToEPredictionProfile, ground_truth : dict, lambda_value : float, alpha : float, beta : float, **kwargs) -> dict: 
    """
    Compute alpha lambda metric, a common metric in prognostics. Alpha-Lambda is met if alpha % of the Time to Event (TtE) distribution is within beta % of the ground truth at prediction time lambda.

    Args:
        toe_profile (ToEPredictionProfile): A profile of predictions, the combination of multiple predictions
        ground_truth (dict): Ground Truth time of event for each event (e.g., {'event1': 748, 'event2', 2233, ...})
        lambda_value (float): Prediction time at or after which metric is evaluated. Evaluation occurs at this time (if a prediction exists) or the next prediction following.
        alpha (float): percentage bounds around time to event (where 0.2 allows 20% error TtE)
        beta (float): portion of prediction that must be within those bounds
        kwargs (optional): configuration arguments. Accepted args include:
            * keys (list[string], optional): list of keys to use. If not provided, all keys are used.

    Returns:
        dict: dictionary containing key value pairs for each key and whether the alpha-lambda was met.
    """
    params = {
        'print': False
    }
    params.update(kwargs)

    for (t_prediction, toe) in toe_profile.items():
        if (t_prediction >= lambda_value):
            # If keys not provided, use all
            keys = params.setdefault('keys', toe.keys())

            bounds = {key : [gt - alpha*(gt-t_prediction), gt + alpha*(gt-t_prediction)] for key, gt in ground_truth.items()}
            pib = toe.percentage_in_bounds(bounds)
            result = {key: pib[key] >= beta for key in keys}
            if params['print']:
                for key in keys:
                    print('\n', key)
                    print('\ttoe:', toe.key(key))
                    print('\tBounds: [{} - {}]({}%)'.format(bounds[key][0], bounds[key][1], pib[key]))
            return result

def prognostic_horizon(toe_profile : ToEPredictionProfile, criteria_eqn : Callable, ground_truth : dict, **kwargs) -> dict:
    """
    Compute prognostic horizon metric, defined as the difference between a time ti, when the predictions meet specified performance criteria, and the time corresponding to the true Time of Event (ToE), for each event.
    PH = ToE - ti
    Args:
        toe_profile (ToEPredictionProfile): A profile of predictions, the combination of multiple predictions
        criteria_eqn (Callable function): A function (tte: UncertainData, ground_truth_tte: dict[str, float]) -> dict[str, bool] calculating whether a prediction in ToEPredictionProfile meets some criteria. \n
            | Args: 
            |  * tte (UncertainData): A single prediction of Time of Event (ToE)
            |  * ground truth tte (dict[str, float]): Ground truth passed into prognostics_horizon
            | Returns: Map of event names to boolean representing if the event has been met. 
            |   e.g., {'event1': True, 'event2': False}
        ground_truth (dict): Dictionary containing ground truth; specified as key, value pairs for event and its value. E.g, {'event1': 47.3, 'event2': 52.1, 'event3': 46.1}
        kwargs (optional): configuration arguments. Accepted args include:
            * print (bool): Boolean specifying whether the prognostic horizon metric should be printed.

    Returns:
        dict: Dictionary containing prognostic horizon calculations (value) for each event (key). e.g., {'event1': 12.3, 'event2': 15.1}
    """
    params = {
        'print': False
    }
    params.update(kwargs)

    ph_result = {k:None for k in ground_truth.keys()} # False means not yet met; will be either a numerical value or None if met
    for (t_prediction, toe) in toe_profile.items():
        # Convert to TtE for toe and ground_truth
        tte = toe - t_prediction
        ground_truth_tte = {}
        ground_truth_tte = {k:v-t_prediction for k,v in ground_truth.items()}
        # Pass to criteria_eqn
        criteria_eqn_dict = criteria_eqn(tte, ground_truth_tte) # -> dict[event_names as str, bool]
        for k,v in criteria_eqn_dict.items():
            if v and (ph_result[k] == None):
                ph_calc = ground_truth[k] - t_prediction
                if ph_calc > 0:
                    ph_result[k] = ph_calc # PH = EOL - ti # ground truth is a dictionary {'EOD': 3005.2} should be ph_result[k] = g_truth[key] - t_prediction
                if (all(v != None for v in ph_result.values())):
                    # Return PH once all criteria are met
                    return ph_result
    # Return PH when criteria not met for at least one event key
    return ph_result

def cumulative_relative_accuracy(toe_profile : ToEPredictionProfile, ground_truth : dict, **kwargs) -> Dict[str, float]:
    """
    Compute cumulative relative accuracy for a given profile, defined as the normalized sum of relative prediction accuracies at specific time instances.
    
    CRA = Σ(RA)/N for each event
    Where Σ is summation of all relative accuracies for a given profile and N is the total count of profiles (Journal Prognostics Health Management, Saxena et al.)
    Args:
        toe_profile (ToEPredictionProfile): A profile of predictions, the combination of multiple predictions
        ground_truth (dict): Dictionary containing ground truth; specified as key, value pairs for event and its value. E.g, {'event1': 47.3, 'event2': 52.1, 'event3': 46.1}
        kwargs (optional): configuration arguments. Accepted args include:

    Returns:
        dict: Dictionary containing cumulative relative accuracy (value) for each event (key). e.g., {'event1': 12.3, 'event2': 15.1}
    """
    ra_sums = defaultdict(int)
    for uncertaindata in toe_profile.values():
        for event,value in uncertaindata.relative_accuracy(ground_truth).items():
            ra_sums[event] += value
    return {event:ra_sum/len(toe_profile) for event, ra_sum in ra_sums.items()}

def monotonicity(toe_profile : ToEPredictionProfile, **kwargs) -> Dict[str, float]:
        """Calculate monotonicty for a prediction profile. 
        Given a prediction profile, for each prediction: go through all predicted events and compare those to the next one.
        Calculates monotonicity for each prediction key using its associated mean value in UncertainData.
        
        monotonoicity = |Σsign(i+1 - i) / N-1|
        Where N is number of measurements and sign indicates sign of calculation.
        Coble, J., et. al. (2021). Identifying Optimal Prognostic Parameters from Data: A Genetic Algorithms Approach. Annual Conference of the PHM Society.
        http://www.papers.phmsociety.org/index.php/phmconf/article/view/1404
        Baptistia, M., et. al. (2022). Relation between prognostics predictor evaluation metrics and local interpretability SHAP values. Aritifical Intelligence, Volume 306.
        https://www.sciencedirect.com/science/article/pii/S0004370222000078

        Args:
            toe_profile (ToEPredictionProfile): A profile of predictions, the combination of multiple predictions
        Returns:
            dict (str, float): Dictionary where keys represent an event and values are float representing its respective monotonicitiy value between [0, 1].
        """
        result = dict()
        by_event = defaultdict(list)
        for time,uncertaindata in toe_profile.items():
            # Collect and organize mean values for each event in the individual prediction v
            for event,value in uncertaindata.mean.items():
                by_event[event].append(value - time)
        # For each event of this prediction v, calculate monotonicity using formula
        for key,l in by_event.items():
            mono_sum = []
            for i in range(len(l)-1): 
                mono_sum.append(sign(l[i+1] - l[i])) 
            result[key] = abs(sum(mono_sum) / (len(l)-1))
        return result



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

from abc import ABC, abstractmethod, abstractproperty
from collections import defaultdict
from matplotlib.figure import Figure
from numpy import array

from ..utils.table import print_table_recursive
from ..visualize import plot_scatter, plot_hist
from prog_models.utils.containers import DictLikeMatrixWrapper


class UncertainData(ABC):
    """
    Abstract base class for data with uncertainty. Any new uncertainty type must implement this class
    """

    def __init__(self, _type = dict):
        self._type = _type
    
    @abstractmethod
    def sample(self, nSamples : int = 1):
        """Generate samples from data

        Args:
            nSamples (int, optional): Number of samples to generate. Defaults to 1.

        Returns:
            samples (UnweightedSamples): Array of nSamples samples

        Example:
            ::
            
                samples = data.samples(100)
        """

    @property
    @abstractproperty
    def median(self) -> dict:
        """The median of the UncertainData distribution or samples 

        Returns:
            dict[str, float]: Median value. e.g., {'key1': 23.2, ...}

        Example:
            ::
            
                median_value = data.median
        """

    @property
    @abstractproperty
    def mean(self) -> dict:
        """The mean of the UncertainData distribution or samples 

        Returns:
            dict[str, float]: Mean value. e.g., {'key1': 23.2, ...}

        Example:
            ::
                
                mean_value = data.mean
        """

    @property
    @abstractproperty
    def cov(self) -> array:
        """The covariance matrix of the UncertiantyData distribution or samples in order of keys (i.e., cov[1][1] is the standard deviation for key keys()[1])

        Returns:
            np.array[np.array[float]]: Covariance matrix

        Example:
            ::

                covariance_matrix = data.cov
        """

    def relative_accuracy(self, ground_truth : dict) -> dict:
        """The relative accuracy is how close the mean of the distribution is to the ground truth, on relative terms
        
        :math:`RA = 1 - \dfrac{\| r-p \|}{r}`

        Where r is ground truth and p is mean of predicted distribution [0]_

        Returns:
            dict[str, float]: Relative accuracy for each event where value is relative accuracy between [0,1]

        Example:
            ::

                ra = data.relative_accuracy({'key1': 22, 'key2': 57})

        References:
            .. [0] Prognostics: The Science of Making Predictions (Goebel et al, 239)
        """
        # if this check isn't here, goes to divide by zero check and raises AttributeError instead of TypeError. Keep? There are unittests checking for type
        if not (isinstance(ground_truth, dict) or isinstance(ground_truth, DictLikeMatrixWrapper)):
            raise TypeError("Ground truth must be passed as a dictionary or *.container argument.")
        if not all(ground_truth.values()):
            raise ZeroDivisionError("Ground truth values must be non-zero in calculating relative accuracy.")
        return {k:1 - (abs(ground_truth[k] - v)/ground_truth[k]) for k,v in self.mean.items()}

    @abstractmethod
    def keys(self):
        """Get the keys for the property represented

        Returns:
            list[str]: keys

        Example:
            ::
                
                keys = data.keys()
        """

    def __contains__(self, key):
        return key in self.keys()

    def percentage_in_bounds(self, bounds : tuple, keys : list = None, n_samples = 1000) -> dict:
        """Calculate percentage of dist is within specified bounds

        Args:
            bounds (tuple[float, float] or dict): Lower and upper bounds. \n
                if tuple: (lower, upper)\n
                if dict: {key: (lower, upper), ...}
            keys (list[str], optional): UncertainData keys to consider when calculating. Defaults to all keys.
            n_samples (int, optional): Number of samples to use when calculating

        Returns:
            dict: Percentage within bounds for each key in keys (where 0.5 = 50%). e.g., {'key1': 1, 'key2': 0.75}

        Example:
            ::
            
                data.percentage_in_bounds((1025, 1075))
                data.percentage_in_bounds({'key1': (1025, 1075), 'key2': (2520, 2675)})
                data.percentage_in_bounds((1025, 1075), keys=['key1', 'key3'])
        """
        return self.sample(n_samples).percentage_in_bounds(bounds, keys=keys)

    def metrics(self, **kwargs) -> dict:
        """Calculate Metrics for this dist

        Keyword Args:
            ground_truth (int or dict, optional): Ground truth value. Defaults to None.
            n_samples (int, optional): Number of samples to use for calculating metrics (if not UnweightedSamples)
            keys (list[str], optional): Keys to calculate metrics for. Defaults to all keys.

        Returns:
            dict: Dictionary of metrics

        Example:
            ::

                print(data.metrics())
                m = data.metrics(ground_truth={'key1': 200, 'key2': 350})
                m = data.metrics(keys=['key1', 'key3'])
        """
        from ..metrics import calc_metrics
        return calc_metrics(self, **kwargs)

    def plot_scatter(self, fig : Figure = None, keys : list = None, num_samples : int = 100, **kwargs) -> Figure:
        """
        Produce a scatter plot

        Args:
            fig (Figure, optional): Existing figure previously used to plot states. If passed a figure argument additional data will be added to the plot. Defaults to creating new figure
            keys (list[str], optional): Keys to plot. Defaults to all keys.
            num_samples (int, optional): Number of samples to plot. Defaults to 100
            **kwargs (optional): Additional keyword arguments passed to scatter function.

        Returns:
            Figure

        Example:
            ::

                m = [5, 7, 3]
                c = [[0.3, 0.5, 0.1], [0.6, 0.7, 1e-9], [1e-9, 1e-10, 1]]
                d = MultivariateNormalDist(['a', 'b', 'c'], m, c)
                d.plot_scatter() # With 100 samples
                states.plot_scatter(num_samples=5) # Specifying the number of samples to plot
                states.plot_scatter(keys=['a', 'b']) # only plot those keys
        """
        if keys is None:
            keys = self.keys()
        samples = self.sample(num_samples)
        return plot_scatter(samples, fig=fig, keys=keys, **kwargs)

    def plot_hist(self, fig = None, keys = None, num_samples = 100, **kwargs):
        """
        Create a histogram

        Args:
            fig (MatPlotLib Figure, optional): Existing histogram figure to be overritten. Defaults to create new figure.
            num_samples (int, optional): Number of samples to plot. Defaults to 100
            keys (list(String), optional): Keys to be plotted. Defaults to None.

        Example:
            ::
                
                m = [5, 7, 3]
                c = [[0.3, 0.5, 0.1], [0.6, 0.7, 1e-9], [1e-9, 1e-10, 1]]
                d = MultivariateNormalDist(['a', 'b', 'c'], m, c)
                d.plot_hist() # With 100 samples
                states.plot_hist(num_samples=20) # Specifying the number of samples to plot
                states.plot_hist(keys=['a', 'b']) # only plot those keys
        """
        if keys is None:
            keys = self.keys()
        samples = self.sample(num_samples)
        return plot_hist(samples, fig=fig, keys=keys, **kwargs)

    def describe(self, title : str = "UncertainData Metrics", print : bool = True) -> defaultdict:
        """
        Print and view basic statistical information about this UncertainData object in a text-based printed table.

        Args:
            title : str
                Title of the table, printed before data rows.
            print : bool = True 
                Optional argument specifying whether to print or not; default true.

        Returns:
            defaultdict
                Dictionary of lists used to print metrics.
        
        Example:
            ::

                data.describe()
        """
        recursive_metrics_table = print_table_recursive(self.metrics(), title, print)
        return recursive_metrics_table

    @abstractmethod
    def __add__(self, other : int) -> "UncertainData":
        """Overriding __add__ (+ operator) for UncertainData. 

        Args:
            other (int): Integer value to be applied to class where appropriate.
        """

    @abstractmethod
    def __radd__(self, other : int) -> "UncertainData":
        """Overriding __radd__ (+ operator right) for UncertainData. 

        Args:
            other (int): Integer value to be applied to class where appropriate.
        """

    @abstractmethod
    def __iadd__(self, other : int) -> "UncertainData":
        """Overriding __iadd__ (+= operator) for UncertainData. 

        Args:
            other (int): Integer value to be applied to class where appropriate.
        """

    @abstractmethod
    def __sub__(self, other : int) -> "UncertainData":
        """Overriding __sub__ (- operator) for UncertainData. 

        Args:
            other (int): Integer value to be applied to class where appropriate.
        """

    @abstractmethod
    def __rsub__(self, other : int) -> "UncertainData":
        """Overriding __rsub__ (- operator right) for UncertainData. 

        Args:
            other (int): Integer value to be applied to class where appropriate.
        """

    @abstractmethod
    def __isub__(self, other : int) -> "UncertainData":
        """Overriding __isub__ (-= operator) for UncertainData. 

        Args:
            other (int): Integer value to be applied to class where appropriate.
        """



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from . import samples
from .toe_metrics import prob_success
from .uncertain_data_metrics import calc_metrics
from .toe_profile_metrics import alpha_lambda, prognostic_horizon, cumulative_relative_accuracy, monotonicity

__all__ = ['alpha_lambda', 'calc_metrics', 'prob_success', 'prognostic_horizon', 'cumulative_relative_accuracy', 'monotonicity']



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from copy import deepcopy
from typing import Callable
from prog_models.sim_result import SimResult, LazySimResult
from prog_algs.uncertain_data import UnweightedSamples, UncertainData

from .prediction import UnweightedSamplesPrediction, PredictionResults
from .predictor import Predictor


class MonteCarlo(Predictor):
    """
    Class for performing a monte-carlo model-based prediction.

    A Predictor using the monte carlo algorithm. The provided initial states are simulated until either a specified time horizon is met, or the threshold for all simulated events is reached for all samples. A provided future loading equation is used to compute the inputs to the system at any given time point. 

    The following configuration parameters are supported (as kwargs in constructor or as parameters in predict method):
    
    Configuration Parameters
    ------------------------------
    t0 : float
        Initial time at which prediction begins, e.g., 0
    dt : float
        Simulation step size (s), e.g., 0.1
    events : list[str]
        Events to predict (subset of model.events) e.g., ['event1', 'event2']
    horizon : float
        Prediction horizon (s)
    n_samples : int
        Number of samples to use. If not specified, a default value is used. If state is type UnweightedSamples and n_samples is not provided, the provided unweighted samples will be used directly.
    save_freq : float
        Frequency at which results are saved (s)
    save_pts : list[float]
        Any additional savepoints (s) e.g., [10.1, 22.5]
    """

    default_parameters = { 
        'n_samples': 100  # Default number of samples to use, if none specified
    }

    def predict(self, state : UncertainData, future_loading_eqn : Callable, **kwargs) -> PredictionResults:
        if isinstance(state, dict) or isinstance(state, self.model.StateContainer):
            from prog_algs.uncertain_data import ScalarData
            state = ScalarData(state, _type = self.model.StateContainer)
        elif isinstance(state, UncertainData):
            state._type = self.model.StateContainer
        else:
            raise TypeError("state must be UncertainData, dict, or StateContainer")

        params = deepcopy(self.parameters) # copy parameters
        params.update(kwargs) # update for specific run
        params['print'] = False
        params['progress'] = False

        if len(params['events']) == 0 and 'horizon' not in params:
            raise ValueError("If specifying no event (i.e., simulate to time), must specify horizon")

        # Sample from state if n_samples specified or state is not UnweightedSamples
        if not isinstance(state, UnweightedSamples) or len(state) != params['n_samples']:
            state = state.sample(params['n_samples'])

        es_eqn = self.model.event_state
        tm_eqn = self.model.threshold_met
        simulate_to_threshold = self.model.simulate_to_threshold

        time_of_event_all = []
        last_states = []
        times_all = []
        inputs_all = []
        states_all = []
        outputs_all = []
        event_states_all = []

        # Perform prediction
        t0 = params.get('t0', 0)
        for x in state:
            first_output = self.model.output(x)
            
            time_of_event = {}
            last_state = {}

            params['t0'] = t0
            params['x'] = x

            if 'save_freq' in params and not isinstance(params['save_freq'], tuple):
                params['save_freq'] = (params['t0'], params['save_freq'])
            
            if len(params['events']) == 0:  # Predict to time
                (times, inputs, states, outputs, event_states) = simulate_to_threshold(future_loading_eqn,       
                    first_output, 
                    threshold_keys = [],
                    **params
                )
            else:
                events_remaining = params['events'].copy()

                times = []
                inputs = SimResult(_copy = False)
                states = SimResult(_copy = False)
                outputs = LazySimResult(fcn = self.model.output, _copy = False)
                event_states = LazySimResult(fcn = es_eqn, _copy = False)

                # Non-vectorized prediction
                while len(events_remaining) > 0:  # Still events to predict
                    (t, u, xi, z, es) = simulate_to_threshold(future_loading_eqn,       
                        first_output, 
                        threshold_keys = events_remaining,
                        **params
                    )

                    # Add results
                    times.extend(t)
                    inputs.extend(u)
                    states.extend(xi)
                    outputs.extend(z, _copy = False)
                    event_states.extend(es, _copy = False)

                    # Get which event occurs
                    t_met = tm_eqn(states[-1])
                    t_met = {key: t_met[key] for key in events_remaining}  # Only look at remaining keys

                    try:
                        event = list(t_met.keys())[list(t_met.values()).index(True)]
                    except ValueError:
                        # no event has occured - hit horizon
                        for event in events_remaining:
                            time_of_event[event] = None
                            last_state[event] = None
                        break

                    # An event has occured
                    time_of_event[event] = times[-1]
                    events_remaining.remove(event)  # No longer an event to predect to

                    # Remove last state (event)
                    params['t0'] = times.pop()
                    inputs.pop()
                    params['x'] = states.pop()
                    last_state[event] = params['x'].copy()
                    outputs.pop()
                    event_states.pop()
            
            # Add to "all" structures
            if len(times) > len(times_all): # Keep longest
                times_all = times
            inputs_all.append(inputs)
            states_all.append(states)
            outputs_all.append(outputs)
            event_states_all.append(event_states)
            time_of_event_all.append(time_of_event)
            last_states.append(last_state)
              
        inputs_all = UnweightedSamplesPrediction(times_all, inputs_all)
        states_all = UnweightedSamplesPrediction(times_all, states_all)
        outputs_all = UnweightedSamplesPrediction(times_all, outputs_all)
        event_states_all = UnweightedSamplesPrediction(times_all, event_states_all)
        time_of_event = UnweightedSamples(time_of_event_all)

        # Transform final states:
        time_of_event.final_state = {
            key: UnweightedSamples([sample[key] for sample in last_states], _type = self.model.StateContainer) for key in time_of_event.keys()
        }

        return PredictionResults(
            times_all, 
            inputs_all, 
            states_all, 
            outputs_all, 
            event_states_all, 
            time_of_event
        )



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from .kalman_filter import KalmanFilter
from .particle_filter import ParticleFilter
from .state_estimator import StateEstimator
from .unscented_kalman_filter import UnscentedKalmanFilter
__all__ = ['KalmanFilter', 'state_estimator', 'StateEstimator', 'unscented_kalman_filter', 'UnscentedKalmanFilter', 'particle_filter', 'ParticleFilter']



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

from collections import UserList
from collections.abc import Iterable
from numpy import array, cov, random
from warnings import warn

from prog_models.utils.containers import DictLikeMatrixWrapper

from . import UncertainData


class UnweightedSamples(UncertainData, UserList):
    """
    Uncertain Data represented by a set of samples. Objects of this class can be treated like a list where samples[n] returns the nth sample (Dict). 

    Args:
        samples (array, dict, or model.*Container, optional): array of samples. Defaults to empty array.\n
            If dict, must be of the form of {key: [value, ...], ...}\n
            If list, must be of the form of [{key: value, ...}, ...]\n
            If InputContainer, OutputContainer, or StateContainer, must be of the form of *Container({'key': value, ...})
    """
    def __init__(self, samples : list = [], _type = dict):
        super().__init__(_type)
        if isinstance(samples, dict) or isinstance(samples, DictLikeMatrixWrapper):
            # Is in form of {key: [value, ...], ...}
            # Convert to array of samples
            if len(samples.keys()) == 0:
                self.data = []  # is empty
                return
            n_samples = len(list(samples.values())[0])  # Number of samples
            self.data = [{key: value[i] for key, value in samples.items()} for i in range(n_samples)]
        elif isinstance(samples, Iterable):
            # is in form of [{key: value, ...}, ...]
            self.data = samples
        else:
            raise ValueError('Invalid input. Must be list or dict, was {}'.format(type(samples)))

    def __eq__(self, other):
        return isinstance(other, UnweightedSamples) and self.data == other.data

    def __getitem__(self,n):
        datem = self.data[n]
        return self._type(datem) if datem is not None else None

    def __add__(self, other : int) -> "UncertainData":
        if other == 0:
            return self
        result = []
        for i in range(len(self.data)):
            new_dict = {}
            for k,v in self.data[i].items():
                new_dict[k] = v + other
            result.append(new_dict)
        return UnweightedSamples(result)

    def __radd__(self, other : int) -> "UncertainData":
        return self.__add__(other)

    def __iadd__(self, other : int) -> "UncertainData":
        if other != 0:
            for i in range(len(self.data)):
                for k,v in self.data[i].items():
                    self.data[i][k] += other
        return self

    def __sub__(self, other : int) -> "UncertainData":
        if other == 0:
            return self
        result = []
        for i in range(len(self.data)):
            new_dict = {}
            for k,v in self.data[i].items():
                new_dict[k] = v - other
            result.append(new_dict)
        return UnweightedSamples(result)

    def __rsub__(self, other : int) -> "UncertainData":
        return self.__sub__(other)

    def __isub__(self, other : int) -> "UncertainData":
        if other != 0:
            for i in range(len(self.data)):
                for k,v in self.data[i].items():
                    self.data[i][k] -= other
        return self

    def __reduce__(self):
        return (UnweightedSamples, (self.data, ))

    def sample(self, num_samples : int = 1, replace = True) -> "UnweightedSamples":
        # Completely random resample
        indices = random.choice(len(self.data), num_samples, replace = replace)
        return UnweightedSamples([self.data[i] for i in indices], _type = self._type)

    def keys(self) -> list:
        if len(self.data) == 0:
            return []  # is empty
        for sample in self:
            if sample is not None:
                return sample.keys()
        return []  # Every element is none

    def key(self, key) -> list:
        """Return samples for given key

        Args:
            key (str): key

        Returns:
            list: list of values for given key
        """
        return [sample[key] for sample in self.data if sample is not None]

    @property
    def median(self) -> dict:
        # Calculate Geometric median of all samples
        min_value = float('inf')
        none_flag = False
        for i, datem in enumerate(self.data):
            if datem is None:
                continue
            p1 = array([d for d in datem.values() if d is not None])
            if not none_flag and len(p1) < len(datem):
                none_flag = True
                warn("Some samples were None, resulting median is of all non-None samples. Note: in some cases, this will bias the median result.")
            total_dist = sum(
                sum((p1 - array([di for di in d.values() if di is not None]))**2)  # Distance between 2 points
                for d in self.data if d is not None)  # For each point
            if total_dist < min_value:
                min_index = i
                min_value = total_dist
        return self._type(self[min_index])

    @property
    def mean(self) -> dict:
        mean = {}
        for key in self.keys():
            values = array([x[key] for x in self.data if x is not None and x[key] is not None])
            if len(values) < len(self.data):
                warn("Some samples were None, resulting mean is of all non-None samples. Note: in some cases, this will bias the mean result.")
            mean[key] = values.mean()
        return self._type(mean)

    @property
    def cov(self) -> dict:
        if len(self.data) == 0:
            return [[]]
        unlabeled_samples = array([[x[key] for x in self.data if x is not None and x[key] is not None] for key in self.keys()])
        if len(unlabeled_samples) < len(self.data):
            warn("Some samples were None, resulting covariance is of all non-None samples. Note: in some cases, this will bias the covariance result.")
        return cov(unlabeled_samples)

    def __str__(self):
        return 'UnweightedSamples({})'.format(self.data)

    @property
    def size(self) -> int:
        """Get the number of samples. Note: kept for backwards compatibility, prefer using len() instead.

        Returns:
            int: Number of samples
        """
        return len(self)

    def percentage_in_bounds(self, bounds, keys : list = None) -> dict:
        if not keys:
            keys = self.keys()
        if isinstance(keys, str):
            keys = [keys]
        if isinstance(bounds, list):
            bounds = {key: bounds for key in self.keys()}
        if not isinstance(bounds, dict) or all([isinstance(b, list) and len(b) == 2 for b in bounds]):
            raise TypeError("Bounds must be list [lower, upper] or dict (key: [lower, upper]), was {}".format(type(bounds)))
        n_elements = len(self.data)
        return {key: sum([x is not None and x < bounds[key][1] and x > bounds[key][0] for x in self.key(key)])/n_elements for key in keys}

    def raw_samples(self):
        warn("raw_samples is deprecated and will be removed in the future.")
        return self.data



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

from collections import UserList, defaultdict, namedtuple
from typing import Dict, List
from numpy import sign
from warnings import warn

from ..uncertain_data import UnweightedSamples, UncertainData

PredictionResults = namedtuple('PredictionResults', ["times", "inputs", "states", "outputs", "event_states", "time_of_event"])


class Prediction():
    """
    Class for the result of a prediction. Is returned by the predict method of a predictor.

    Args:
        times (list[float]):
            Times for each data point where times[n] corresponds to data[n]
        data (list[UncertainData]):
            Data points for each time in times 
    """

    def __init__(self, times : list, data : list):
        self.times = times
        self.data = data

    def __eq__(self, other : "Prediction") -> bool:
        """Compare 2 Predictions

        Args:
            other (Prediction):

        Returns:
            bool: If the two Predictions are equal
        """
        return self.times == other.times and self.data == other.data

    def snapshot(self, time_index : int) -> UncertainData:
        """Get all samples from a specific timestep

        Args:
            index (int):
                Timestep (index number from times)

        Returns:
            UncertainData: Distribution for time corresponding to times[timestep]
        """
        return self.data[time_index]

    @property
    def mean(self) -> List[dict]:
        """Estimate of the mean value of the prediction at each time

        Returns:
            list[dict]: 
                Mean value of the prediction at each time where mean[n] corresponds to the mean value of the prediction at time times[n].\n
                The mean value at each time is a dictionary. \n
                e.g., [{'state1': 1.2, 'state2': 1.3, ...}, ...]

        Example:
            mean_value = data.mean
        """
        return [dist.mean for dist in self.data]

    def time(self, index : int):
        warn("Deprecated. Please use prediction.times[index] instead.")
        return self.times[index]

    def monotonicity(self) -> Dict[str, float]:
        """Calculate monotonicty for a single prediction. 
        Given a single prediction, for each event: go through all predicted states and compare those to the next one.
        Calculates monotonicity for each event key using its associated mean value in UncertainData.
        
        :math:`monotonoicity = \| \Sigma \dfrac{sign(i+1 - i)}{N-1}\|`

        Where N is number of measurements and sign indicates sign of calculation [0]_ [1]_.

        Returns:
            dict (str, float): Value between [0, 1] indicating monotonicity of a given event for the Prediction.

        References:
            .. [0] Coble, J., et. al. (2021). Identifying Optimal Prognostic Parameters from Data: A Genetic Algorithms Approach. Annual Conference of the PHM Society. http://www.papers.phmsociety.org/index.php/phmconf/article/view/1404
            .. [1] Baptistia, M., et. al. (2022). Relation between prognostics predictor evaluation metrics and local interpretability SHAP values. Aritifical Intelligence, Volume 306. https://www.sciencedirect.com/science/article/pii/S0004370222000078

        """
        # Collect and organize mean values for each event
        by_event = defaultdict(list)
        for uncertaindata in self.data:
            for key,value in uncertaindata.mean.items():
                by_event[key].append(value)

        # For each event, calculate monotonicity using formula
        result = {}
        for key,l in by_event.items():
            mono_sum = []
            for i in range(len(l)-1): 
                mono_sum.append(sign(l[i+1] - l[i])) 
            result[key] = abs(sum(mono_sum) / (len(l)-1))
        return result

class UnweightedSamplesPrediction(Prediction, UserList):
    """
    Immutable data class for the result of a prediction, where the predictions are stored as UnweightedSamples. Is returned from the predict method of a sample based prediction class (e.g., MonteCarlo). Objects of this class can be iterated and accessed like a list (e.g., prediction[0]), where prediction[n] represents a profile for sample n.

    Args:
        times (list[float]):
            Times for each data point where times[n] corresponds to data[:][n]
        data (list[SimResult]):
            Data points where data[n] is a SimResult for sample n
    """

    def __init__(self, times : list, data : list):
        super(UnweightedSamplesPrediction, self).__init__(times, data)
        self.__transformed = False  # If transform has been calculated

    def __calculate_tranform(self):
        """
        Calculate tranform of the data from data[sample_id][time_id] to data[time_id][sample_id]. Result is cached as self.__transform and is used in methods which look at a snapshot for a specific time
        """
        # Lazy calculation of tranform - only if needed
        # Note: prediction stops when event is reached, so for the length of all will not be the same. 
        # If the prediction doesn't go this far, then the value is set to None
        self.__transform = [UnweightedSamples([sample[time_index] if len(sample) > time_index else None for sample in self.data]) for time_index in range(len(self.times))]
        self.__transformed = True

    def __str__(self) -> str:
        return "UnweightedSamplesPrediction with {} savepoints".format(len(self.times))

    @property
    def mean(self) -> list:
        if not self.__transformed:
            self.__calculate_tranform()
        return [dist.mean for dist in self.__transform]

    def sample(self, sample_id : int):
        warn("Deprecated. Please use prediction[sample_id] instead.")
        return self[sample_id]

    def snapshot(self, time_index : int) -> UnweightedSamples:
        """Get all samples from a specific timestep

        Args:
            index (int): Timestep (index number from times)

        Returns:
            UnweightedSamples: Samples for time corresponding to times[timestep]
        """
        if not self.__transformed:
            self.__calculate_tranform()
        return self.__transform[time_index]

    def __not_implemented(self, *args, **kw):
        """
        This function is not implemented. Calling this will raise an error. Is is only included to make the class immutable.

        Raises:
            ValueError: 
        """
        raise ValueError("UnweightedSamplesPrediction is immutable (i.e., read only)")

    append = __not_implemented
    extend = __not_implemented
    clear = __not_implemented
    reverse = __not_implemented
    remove = __not_implemented
    insert = __not_implemented
    pop = __not_implemented
    __setitem__ = __not_implemented
    __setslice__ = __not_implemented
    __delitem__ = __not_implemented



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from abc import ABC, abstractmethod, abstractproperty
from copy import deepcopy

from ..uncertain_data import UncertainData
from ..exceptions import ProgAlgTypeError


class StateEstimator(ABC):
    """
    Interface class for state estimators

    Abstract base class for creating state estimators that perform state estimation. Subclasses must implement this interface. Equivilant to "Observers" in NASA's Matlab Prognostics Algorithm Library

    Args:
        model (PrognosticsModel):
            A prognostics model to be used in state estimation
            See: Prognostics Model Package
        x0 (UncertainData, model.StateContainer, or dict):
            Initial (starting) state, with keys defined by model.states \n
            e.g., x = ScalarData({'abc': 332.1, 'def': 221.003}) given states = ['abc', 'def']
    
    See state-estimator specific documentation for speicfic keyword arguments.
    """

    default_parameters = {
        't0': 0
    }

    def __init__(self, model, x0, **kwargs):
        # Check model
        if not hasattr(model, 'output'):
            raise ProgAlgTypeError("model must have `output` method")
        if not hasattr(model, 'next_state'):
            raise ProgAlgTypeError("model must have `next_state` method")
        if not hasattr(model, 'outputs'):
            raise ProgAlgTypeError("model must have `outputs` property")
        if not hasattr(model, 'states'):
            raise ProgAlgTypeError("model must have `states` property")
        self.model = model

        # Check x0
        for key in model.states:
            if key not in x0:
                raise ProgAlgTypeError("x0 missing state `{}`".format(key))
        
        # Process kwargs (configuration)
        self.parameters = deepcopy(self.default_parameters)
        self.parameters.update(kwargs)

        self.t = self.parameters['t0']  # Initial Time

    @abstractmethod
    def estimate(self, t : float, u, z, **kwargs) -> None:
        """
        Perform one state estimation step (i.e., update the state estimate, filt.x)

        Parameters
        ----------
        t : float
            Current timestamp in seconds (≥ 0.0)
            e.g., t = 3.4
        u : InputContainer
            Measured inputs, with keys defined by model.inputs.
            e.g., u = m.InputContainer({'i':3.2}) given inputs = ['i']
        z : OutputContainer
            Measured outputs, with keys defined by model.outputs.
            e.g., z = m.OutputContainer({'t':12.4, 'v':3.3}) given outputs = ['t', 'v']

        Note
        ----
        This method updates the state estimate stored in filt.x, but doesn't return the updated estimate. Call filt.x to get the updated estimate.
        """

    @property
    @abstractproperty
    def x(self) -> UncertainData:
        """
        The current estimated state. 

        Example
        -------
        state = filt.x
        """



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

from typing import Union
from numpy import array

from . import UncertainData, UnweightedSamples


class ScalarData(UncertainData):
    """
    Data without uncertainty- single value

    Args:
        state (dict or Container): Single state in the form of dict or model.*Container (InputContainer, OutputContainer, Statecontainer) representing states and respective values.
    """
    def __init__(self, state, _type = dict): 
        self.__state = state
        super().__init__(_type)
    
    def __reduce__(self):
        return (ScalarData, (self.__state, ))

    def __eq__(self, other : "ScalarData") -> bool:
        return isinstance(other, ScalarData) and other.mean == self.__state

    def __add__(self, other : int) -> "UncertainData":
        if other == 0:
            return self
        new_state = dict()
        for k,v in self.__state.items():
            new_state[k] = v + other
        return ScalarData(new_state)

    def __radd__(self, other : int) -> "UncertainData":
        return self.__add__(other)

    def __iadd__(self, other : int) -> "UncertainData":
        if other != 0:
            for k in self.__state.keys():
                self.__state[k] += other
        return self

    def __sub__(self, other : int) -> "UncertainData":
        new_state = dict()
        for k,v in self.__state.items():
            new_state[k] = v - other
        return ScalarData(new_state)

    def __rsub__(self, other : int) -> "UncertainData":
        return self.__sub__(other)

    def __isub__(self, other : int) -> "UncertainData":
        if other != 0:
            for k in self.__state.keys():
                self.__state[k] -= other
        return self

    @property
    def median(self) -> dict:
        return self.mean
        
    @property
    def mean(self) -> dict:
        return self._type(self.__state)

    @property
    def cov(self) -> array:
        return [[0 for _ in range(len(self.__state))] for _ in range(len(self.__state))]

    def keys(self):
        return self.__state.keys()
        
    def sample(self, num_samples : int = 1) -> UnweightedSamples:
        return UnweightedSamples([self.__state] * num_samples, _type = self._type)

    def __str__(self) -> str:
        return 'ScalarData({})'.format(self.__state)

    def percentage_in_bounds(self, bounds : Union[list, dict]) -> dict:
        if isinstance(bounds, list):
            bounds = {key: bounds for key in self.keys()}
        if not isinstance(bounds, dict) and all([isinstance(b, list) for b in bounds]):
            raise TypeError("Bounds must be list [lower, upper] or dict (key: [lower, upper]), was {}".format(type(bounds)))
        return {key: (1 if bounds[key][0] < x and bounds[key][1] > x else 0) for (key, x) in self.__state.items()}



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Callable

from prog_algs.predictors.prediction import PredictionResults
from ..exceptions import ProgAlgTypeError
from ..uncertain_data import UncertainData


class Predictor(ABC):
    """
    Interface class for predictors

    Abstract base class for creating predictors that perform prediction. Predictor subclasses must implement this interface. Equivilant to "Predictors" in NASA's Matlab Prognostics Algorithm Library

    Parameters
    ----------
    model : PrognosticsModel
        See: :py:mod:`prog_models` package\n
        A prognostics model to be used in prediction
    kwargs : optional, keyword arguments
    """
    default_parameters = {}

    def __init__(self, model, **kwargs):
        if not hasattr(model, 'output'):
            raise ProgAlgTypeError("model must have `output` method")
        if not hasattr(model, 'next_state'):
            raise ProgAlgTypeError("model must have `next_state` method")
        if not hasattr(model, 'inputs'):
            raise ProgAlgTypeError("model must have `inputs` property")
        if not hasattr(model, 'outputs'):
            raise ProgAlgTypeError("model must have `outputs` property")
        if not hasattr(model, 'states'):
            raise ProgAlgTypeError("model must have `states` property")
        if not hasattr(model, 'simulate_to_threshold'):
            raise ProgAlgTypeError("model must have `simulate_to_threshold` property")
        self.model = model

        self.parameters = deepcopy(self.default_parameters)
        self.parameters['events'] = self.model.events.copy()  # Events to predict to
        self.parameters.update(kwargs)

    @abstractmethod
    def predict(self, state : UncertainData, future_loading_eqn : Callable, **kwargs) -> PredictionResults:
        """
        Perform a single prediction

        Parameters
        ----------
        state : UncertainData 
            Distribution representing current state of the system
        future_loading_eqn : function (t, x) -> z
            Function to generate an estimate of loading at future time t, and state x

        Keyword Arguments
        -------------------
        dt : float
            Simulation step size (s), e.g., 0.1
        events : list[str]
            Events to predict (subset of model.events) e.g., ['event1', 'event2']
        horizon : float
            Prediction horizon (s)
        save_freq : float
            Frequency at which results are saved (s)
        save_pts : list[float]
            Any additional savepoints (s) e.g., [10.1, 22.5]

        Return
        ----------
        NamedTuple
            * times (list[float]): Times for each savepoint such that inputs.snapshot(i), states.snapshot(i), outputs.snapshot(i), and event_states.snapshot(i) are all at times[i]            
            * inputs (:py:class:`Prediction`): Inputs at each savepoint such that inputs.snapshot(i) is the input distribution (type UncertainData) at times[i]
            * states (:py:class:`Prediction`): States at each savepoint such that states.snapshot(i) is the state distribution (type UncertainData) at times[i]
            * outputs (:py:class:`Prediction`): Outputs at each savepoint such that outputs.snapshot(i) is the output distribution (type UncertainData) at times[i]
            * event_states (:py:class:`Prediction`): Event states at each savepoint such that event_states.snapshot(i) is the event state distribution (type UncertainData) at times[i]
            * time_of_event (:py:class:`prog_algs.uncertain_data.UncertainData`): Distribution of predicted Time of Event (ToE) for each predicted event, represented by some subclass of UncertaintData (e.g., :py:class:`MultivariateNormalDist`)
        """
        pass



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

from numpy import array
from numpy.random import multivariate_normal

from . import UncertainData, UnweightedSamples


class MultivariateNormalDist(UncertainData):
    """
    Data represented by a multivariate normal distribution with mean and covariance matrix

    Args:
            labels (list[str]): Labels for states, in order of mean values
            mean (array[float]): Mean values for state in the same order as labels
            covar (array[array[float]]): Covariance matrix for state
    """
    def __init__(self, labels, mean: array, covar : array, _type = dict):
        self.__labels = list(labels)
        self.__mean = array(list(mean))
        self.__covar = array(list(covar))
        super().__init__(_type)

    def __reduce__(self):
        return (MultivariateNormalDist, (self.__labels, self.__mean, self.__covar))

    def __eq__(self, other : "MultivariateNormalDist") -> bool:
        return isinstance(other, MultivariateNormalDist) and self.keys() == other.keys() and self.mean == other.mean and (self.cov == other.cov).all()

    def __add__(self, other : int) -> "UncertainData":
        if other == 0:
            return self
        return MultivariateNormalDist(self.__labels, array([i+other for i in self.__mean]), self.__covar)

    def __radd__(self, other : int) -> "UncertainData":
        return self.__add__(other)

    def __iadd__(self, other : int) -> "UncertainData":
        if not isinstance(other, (int, float)):
            raise TypeError(f" unsupported operand type(s) for +: '{type(other)}' and '{type(self.__mean[0])}'")
        if other != 0:
            self.__mean = array([i+other for i in self.__mean])
        return self

    def __sub__(self, other : int) -> "UncertainData":
        if other == 0:
            return self
        return MultivariateNormalDist(self.__labels, array([i-other for i in self.__mean]), self.__covar)

    def __rsub__(self, other : int) -> "UncertainData":
        return self.__sub__(other)

    def __isub__(self, other : int) -> "UncertainData":
        if not isinstance(other, (int, float)):
            raise TypeError(f" unsupported operand type(s) for -: '{type(other)}' and '{type(self.__mean[0])}'")
        if other != 0:
            self.__mean = array([i-other for i in self.__mean])
        return self

    def sample(self, num_samples : int = 1) -> UnweightedSamples:
        if len(self.__mean) != len(self.__labels):
            raise Exception("labels must be provided for each value")
    
        samples = multivariate_normal(self.__mean, self.__covar, num_samples)
        samples = [{key: value for (key, value) in zip(self.__labels, x)} for x in samples]
        return UnweightedSamples(samples, _type = self._type)

    def keys(self) -> list:
        return self.__labels

    @property
    def median(self) -> float:
        # For normal distribution medain = mean
        return self.mean

    @property
    def mean(self) -> dict:
        return self._type({key: value for (key, value) in zip(self.__labels, self.__mean)})

    def __str__(self) -> str:
        return 'MultivariateNormalDist(mean: {}, covar: {})'.format(self.__mean, self.__covar)     

    @property
    def cov(self) -> array:
        return self.__covar



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

"""
This file includes functions for calculating metrics specific to Time of Event (ToE) from a single event or multiple events given the same time of prediction
"""
from typing import Iterable
from numpy import isscalar

from ..uncertain_data import UncertainData, UnweightedSamples

def prob_success(toe : UncertainData, time : float, **kwargs) -> float:
    """Calculate probability of success - i.e., probability that event will not occur within a given time (i.e., success)

    Args:
        toe (UncertainData or list[float]): Times of event for a single event (array[float]) or multiple events, output from predictor
        time (float): time for calculation
        **kwargs (optional): Configuration parameters. Supported parameters include:
          * n_samples (int): Number of samples to use for calculating metrics (if ToE is not UnweightedSamples). Defaults to 10,000.
          * keys (list of strings, optional): Keys to calculate metrics for. Defaults to all keys.

    Returns:
        float: Probability of success

    Example:
        ::

            # ToE estimate distribution is returned from a predictor's predict method
            result = predictor.predict(...)
            toe = result.time_of_event 
            
            from prog_algs.metrics import prob_success
            now = 10 # Current time
            p_success = prob_success(toe, now)
            p_success = prob_success(toe, now, n_samples = 100) # Can also specify number of samples
            p_success = prob_success(toe, now, keys = ['event1']) # Can specify specific keys to consider
    """
    params = {
        'n_samples': 10000,  # Default 
    }
    params.update(kwargs)

    if isinstance(toe, UncertainData):
        # Default to all keys
        keys = params.setdefault('keys', toe.keys())
        if isinstance(keys, str):
            keys = [keys]

        if isinstance(toe, UnweightedSamples):
            samples = toe
        else:
            # Some other distribution besides unweighted samples
            # Generate Samples
            samples = toe.sample(params['n_samples'])

        # If unweighted_samples, calculate metrics for each key
        return {key: prob_success(samples.key(key), 
                time, 
                **kwargs) for key in keys}
    elif isinstance(toe, Iterable):
        if len(toe) == 0:
            raise ValueError('Time of Event must not be empty')
        # Is list or array
        if isscalar(toe[0]) or toe[0] is None:
            # list of numbers - this is the case that we can calculate
            pass
        elif isinstance(toe[0], dict):
            # list of dicts - Supported for backwards compatabilities
            toe = UnweightedSamples(toe)
            return prob_success(toe, time, **kwargs)
        else:
            raise TypeError("ToE must be type Uncertain Data or array of dicts, was {}".format(type(toe)))
    else:
        raise TypeError("ToE must be type Uncertain Data or array of dicts, was {}".format(type(toe)))

    return sum([e is None or e > time for e in toe])/len(toe)



# Copyright © 2021 United States Government as represented by the Administrator of the
# National Aeronautics and Space Administration.  All Rights Reserved.

from .table import print_table_recursive




# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from copy import deepcopy
from filterpy import kalman
from numpy import diag, array, transpose, isnan
from typing import Callable

from .prediction import Prediction, UnweightedSamplesPrediction, PredictionResults
from .predictor import Predictor
from prog_algs.uncertain_data import MultivariateNormalDist, UncertainData, ScalarData


class LazyUTPrediction(Prediction):
    def __init__(self, state_prediction, sigma_fcn : Callable, ut_fcn : Callable, transform_fcn : Callable):
        self.times = state_prediction.times
        self.__states = state_prediction
        self.__data = None
        self.__sigma_fcn = sigma_fcn
        self.__transform = transform_fcn
        self.__ut_fcn = ut_fcn

    @property
    def data(self):
        if self.__data == None:
            self.__data = []
            # For each timepoint
            for i in range(len(self.times)):
                x = self.__states.snapshot(i)

                # Get Sigma points
                keys = x.keys()
                mean = [x.mean[key] for key in keys]  # Maintain ordering
                covar = x.cov
                sigma_pts = self.__sigma_fcn.sigma_points(mean, covar)
                
                # Apply Tranformation (e.g., output, event_state)
                sigma_pt_tranformed = [
                    self.__transform({key: value for key, value in zip(keys, sigma_pt)})
                    for sigma_pt in sigma_pts
                ]
                # result is [sigma_pt][ -> output/event_state (dict)]

                transformed_keys = sigma_pt_tranformed[0].keys()

                # Flatten 
                sigma_pt_tranformed = array([array(list(sigma_pt.values())) for sigma_pt in sigma_pt_tranformed]) # map -> array

                # Apply Unscented Transform to form output distribution
                mean, cov = self.__ut_fcn(sigma_pt_tranformed, self.__sigma_fcn.Wm, self.__sigma_fcn.Wc)
                self.__data.append(MultivariateNormalDist(transformed_keys, mean, cov))

        return self.__data


class UnscentedTransformPredictor(Predictor):
    """
    Class for performing model-based prediction using an unscented transform. 

    This class defines logic for performing model-based state prediction using sigma points and an unscented transform. The Unscented Transform Predictor propagates the sigma-points in the state-space in time domain until the event threshold is met. The step at which the i-th sigma point reaches the threshold is the step at which the i-th sigma point will be placed along the time dimension. By repeating the procedure for all sigma-points, we obtain the sigma-points defining the distribution of the time of event (ToE); for example, the End Of Life (EOL) event. The provided future loading equation is used to compute the inputs to the system at any given time point. 

    The following configuration parameters are supported (as kwargs in constructor or as parameters in predict method):

    Configuration Parameters
    ------------------------------
    alpha, beta, kappa: float
        UKF Scaling parameters. See: https://en.wikipedia.org/wiki/Kalman_filter#Unscented_Kalman_filter
    Q: np.array
        Process noise covariance matrix [nStates x nStates]
    t0 : float
        Initial time at which prediction begins, e.g., 0
    dt : float
        Simulation step size (s), e.g., 0.1
    events : list[str]
        Events to predict (subset of model.events) e.g., ['event1', 'event2']
    horizon : float
        Prediction horizon (s)
    save_freq : float
        Frequency at which results are saved (s)
    save_pts : list[float]
        Any additional savepoints (s) e.g., [10.1, 22.5]

    Note
    ----    
    The resulting sigma-points along the time dimension are used to compute mean and covariance of the event time (ToE), under the hypothesis that the ToE distribution would also be well represented by a Gaussian. This is a strong assumption that likely cannot be satisfied for real systems with strong non-linear state propagation or nonlinear ToE curves. Therefore, the user should be cautious and verify that modeling the event time using a Gaussian distribution is satisfactory.
    """
    default_parameters = {
        'alpha': 1,
        'beta': 0,
        'kappa': -1,
        't0': 0,
        'dt': 0.5,
        'horizon': 1e99,
        'save_pts': [],
        'save_freq': 1e99
    }

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

        self.model = model
        self.__input = None  # Input at an individual step. Note, this needs to be a member to pass between state_transition and predict

        # setup UKF
        num_states = model.n_states
        num_measurements = model.n_outputs

        if 'Q' not in self.parameters:
            # Default 
            self.parameters['Q'] = diag([1.0e-8 for _ in range(num_states)])
        
        def measure(x):
            x = model.StateContainer(x)
            z = model.output(x)
            return model.OutputContainer(z)

        def state_transition(x, dt):
            x = model.StateContainer(x)
            x = model.next_state(x, self.__input, dt)
            x = model.apply_limits(x)
            return array(list(x.values()))

        self.sigma_points = kalman.MerweScaledSigmaPoints(num_states, alpha=self.parameters['alpha'], beta=self.parameters['beta'], kappa=self.parameters['kappa'])
        self.filter = kalman.UnscentedKalmanFilter(num_states, num_measurements, self.parameters['dt'], measure, state_transition, self.sigma_points)
        self.filter.Q = self.parameters['Q']

    def predict(self, state, future_loading_eqn : Callable, **kwargs) -> PredictionResults:
        """
        Perform a single prediction

        Parameters
        ----------
        state (UncertaintData): Distribution of states
        future_loading_eqn : function (t, x={}) -> z
            Function to generate an estimate of loading at future time t
        options (optional, kwargs): configuration options\n
        Any additional configuration values. Note: These parameters can also be specified in the predictor constructor. The following configuration parameters are supported: \n
            * alpha, beta, kappa: UKF Scaling parameters
            * t0: Starting time (s)
            * dt : Step size (s)
            * horizon : Prediction horizon (s)
            * events : List of events to be predicted (subset of model.events, default is all events)

        Returns (PredictionResults)
        -------
        times: [number]
            Times for each simulated point in format times[index]
        inputs: [[dict]]
            Future input (from future_loading_eqn) for each sample and time in times
            where inputs[sample_id][index] corresponds to time times[sample_id][index]
        states: [[dict]]
            Estimated states for each sample and time in times
            where states[sample_id][index] corresponds to time times[sample_id][index]
        outputs: [[dict]]
            Estimated outputs for each sample and time in times
            where outputs[sample_id][index] corresponds to time times[sample_id][index]
        event_states: [[dict]]
            Estimated event state (e.g., SOH), between 1-0 where 0 is event occurance, for each sample and time in times
            where event_states[sample_id][index] corresponds to time times[sample_id][index]
        time_of_event: UncertainData
            Estimated time where a predicted event will occur for each sample. Note: Mean and Covariance Matrix will both 
            be nan if every sigma point doesnt reach threshold within horizon
            Also, includes member final_state (time_of_event.final_state) which is the state at the last time step. 
            time_of_event.final_state is a dict of the form {'state_name': state_value}, is equal to None if event does not occur within horizon
        """
        if isinstance(state, dict) or isinstance(state, self.model.StateContainer) or isinstance(state, ScalarData):
            raise TypeError("state must be a distribution (e.g., MultivariateNormalDist, UnweightedSamples), not scalar")
        elif isinstance(state, UncertainData):
            state._type = self.model.StateContainer
        else:
            raise TypeError("state must be UncertainData, dict, or StateContainer")

        params = deepcopy(self.parameters) # copy parameters
        params.update(kwargs) # update for specific run

        if len(params['events']) == 0 and 'horizon' not in params:
            raise ValueError("If specifying no event (i.e., simulate to time), must specify horizon")

        # Optimizations 
        events_to_predict = params['events']
        dt = params['dt']
        model = self.model
        filt = self.filter
        sigma_points = self.sigma_points
        n_points = sigma_points.num_sigmas()
        threshold_met = model.threshold_met
        StateContainer = model.StateContainer

        # Update State 
        self.__state_keys = state_keys = state.mean.keys()  # Used to maintain ordering as we strip keys and return
        filt.x = [x for x in state.mean.values()]
        filt.P = state.cov

        # Setup first states
        t = params['t0']
        save_pt_index = 0
        ToE = {key: [float('nan') for i in range(n_points)] for key in events_to_predict}  # Keep track of final ToE values
        last_state = {key: [None for i in range(n_points)] for key in events_to_predict}  # Keep track of final state values

        times = []
        inputs = []
        states = []
        save_freq = params['save_freq']
        next_save = t + save_freq
        save_pts = params['save_pts']
        save_pts.append(1e99)  # Add last endpoint
        def update_all():
            times.append(t)
            inputs.append(deepcopy(self.__input))  # Avoid optimization where u is not copied
            x_dict = MultivariateNormalDist(self.__state_keys, filt.x, filt.P, _type = self.model.StateContainer)
            states.append(x_dict)  # Avoid optimization where x is not copied

        # Simulation
        self.__input = future_loading_eqn(t, state.mean)
        update_all()  # First State
        while t < params['horizon']:
            # Iterate through time
            t += dt
            mean_state = StateContainer({key: x for (key, x) in zip(state_keys, filt.x)})
            self.__input = future_loading_eqn(t, mean_state)
            filt.predict(dt=dt)

            # Record States
            if (t >= next_save):
                next_save += save_freq
                update_all()
            if (t >= save_pts[save_pt_index]):
                save_pt_index += 1
                update_all()
            
            # Check that any sigma point has hit event
            points = sigma_points.sigma_points(filt.x, filt.P)
            all_failed = True
            for i, point in zip(range(n_points), points):
                # x = StateContainer({key: x for (key, x) in zip(state_keys, point)})
                x = StateContainer(point)
                t_met = threshold_met(x)

                # Check Thresholds
                for key in events_to_predict:
                    if t_met[key]:
                        if isnan(ToE[key][i]):
                            # First time event has been reached
                            ToE[key][i] = t
                            last_state[key][i] = x.copy()
                    else:
                        all_failed = False  # This event for this sigma point hasn't been met yet
            if all_failed:
                # If all events have been reched for every sigma point
                break 
        
        # Prepare Results
        pts = array([[e for e in ToE[key]] for key in ToE.keys()])
        pts = transpose(pts)
        mean, cov = kalman.unscented_transform(pts, sigma_points.Wm, sigma_points.Wc)

        # Transform final state into {event_name: MultivariateNormalDist}
        final_state = {}
        for event_key in last_state.keys():
            if any([last_state_i is None for last_state_i in last_state[event_key]]):
                # If any sigma point has not met the event threshold
                final_state[event_key] = None
                continue
            last_state_pts = array([[last_state_i[state_key] for state_key in state_keys] for last_state_i in last_state[event_key]])
            # last_state_pts = transpose(last_state_pts)
            last_state_mean, last_state_cov = kalman.unscented_transform(last_state_pts, sigma_points.Wm, sigma_points.Wc)
            final_state[event_key] = MultivariateNormalDist(state_keys, last_state_mean, last_state_cov, _type = self.model.StateContainer)

        # At this point only time of event, inputs, and state are calculated 
        inputs_prediction = UnweightedSamplesPrediction(times, [inputs])
        state_prediction = Prediction(times, states)
        output_prediction = LazyUTPrediction(state_prediction, sigma_points, kalman.unscented_transform, model.output)
        event_state_prediction = LazyUTPrediction(state_prediction, sigma_points, kalman.unscented_transform, model.event_state)
        time_of_event = MultivariateNormalDist(ToE.keys(), mean, cov)
        time_of_event.final_state = final_state
        return PredictionResults(
            times, 
            inputs_prediction, 
            state_prediction, 
            output_prediction, 
            event_state_prediction, 
            time_of_event
        )  
      


# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.
from matplotlib.collections import PathCollection
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from math import sqrt
from typing import List

def plot_scatter(samples : List[dict], fig : Figure = None, keys : List[str] = None, legend : str = 'auto', **kwargs) -> Figure:
    """
    Produce a scatter plot for a given list of states

    Args:
        samples ([dict]): Non-empty list of states where each element is a dictionary containing a single sample
        fig (Figure, optional): Existing figure previously used to plot states. If passed a figure argument additional data will be added to the plot. Defaults to creating new figure
        keys (list of strings, optional): Keys to plot. Defaults to all keys.
        legend (optional): When the legend should be shown, options:
            False: Dont show legend
            "auto": Show legend automatically if more than one data set
            True: Always show legend
        **kwargs (optional): Additional keyword arguments passed to scatter function. Includes those supported by scatter

    Returns:
        Figure
    
    Example:
            states = UnweightedSamples([1, 2, 3, 4, 5])
            plot_scatter(states.sample(100)) # With 100 samples
            plot_scatter(states.sample(100), keys=['state1', 'state2']) # only plot those keys
    """
    # Input checks
    if len(samples) <= 0:
        raise Exception('Must include atleast one sample to plot')
    
    if keys is not None:
        try:
            iter(keys)
        except TypeError:
            raise TypeError("Keys should be a list of strings (e.g., ['state1', 'state2'], was {}".format(type(keys)))

    # Handle input
    parameters = {  # defaults
        'alpha': 0.5
    }
    parameters.update(kwargs)

    if keys is None:
        keys = samples[0].keys()
    keys = list(keys)

    n = len(keys)
    if n < 2:
        raise Exception("At least 2 states required for scatter, got {}".format(n))

    if fig is None:
        # If no figure provided, create one
        fig = plt.figure()
        axes = [[fig.add_subplot(n-1, n-1, 1 + i + j*(n-1)) for i in range(n-1)] for j in range(n-1)]
    else:
        # Check size of axes
        if len(fig.axes) != (n-1)*(n-1):
            raise Exception("Cannot use existing figure - Existing figure graphs {} states, data includes {} states".format(sqrt(len(fig.axes))+1, n))

        # Unpack axes
        axes = [[fig.axes[i + j*(n-1)] for i in range(n-1)] for j in range(n-1)]

    for i in range(n-1):
        # For each column
        x_key = keys[i] 

        # Set labels on extremes
        axes[-1][i].set_xlabel(x_key)  # Bottom row
        axes[i][0].set_ylabel(keys[i+1])  # Left column

        # plot 
        for j in range(i, n-1): 
            # for each row
            y_key = keys[j+1]
            x1 = [x[x_key] for x in samples if x is not None]
            x2 = [x[y_key] for x in samples if x is not None]
            axes[j][i].scatter(x1, x2, **parameters)

        # Hide axes not used in plots 
        for j in range(0, i):
            axes[j][i].set_visible(False)

    # Set legend
    if legend == 'auto' or legend:
        labels = [thing.get_label() for thing in axes[0][0].get_children()
            if isinstance(thing, PathCollection)]
        if legend == 'auto' and len(labels) > 0 or legend:
            fig.legend().remove()  # Remove any existing legend - prevents "ghost effect"
            fig.legend(labels=labels, loc='upper right')

    return fig



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

# This file is kept for backwards compatability
from numpy import mean, sqrt
from warnings import warn

from .uncertain_data_metrics import calc_metrics as eol_metrics
from .toe_metrics import prob_success
from .toe_profile_metrics import alpha_lambda

def mean_square_error(values : list, ground_truth : float) -> float:
    """Mean Square Error
    Args:
        values (List[float]): Times of Event (ToE) for a single event, output from predictor
        ground_truth (float): Ground truth ToE
    Returns:
        float: mean square error of ToE predictions
    """
    return sum([(mean(x) - ground_truth)**2 for x in values])/len(values)

def root_mean_square_error(values, ground_truth):
    """Root Mean Square Error
    Args:
        values (List[float]): Times of Event (ToE) for a single event, output from predictor
        ground_truth (float): Ground truth ToE
    Returns:
        float: root mean square error of ToE predictions
    """
    return sqrt(sum([(mean(x) - ground_truth)**2 for x in values])/len(values))

def percentage_in_bounds(toe : list, bounds : tuple) -> float:
    """Calculate percentage of ToE dist is within specified bounds

    Args:
        toe (List[float]): Times of Event (ToE) for a single event, output from predictor
        bounds ((float, float)): Lower and upper bounds

    Returns:
        float: Percentage within bounds (where 1 = 100%)
    """
    warn('percentage_in_bounds has been deprecated in favor of UncertainData.percentage_in_bounds(bounds). This function will be removed in a future release')
    return sum([x < bounds[1] and x > bounds[0] for x in toe])/ len(toe)



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from copy import deepcopy
from filterpy import kalman
import numpy as np
from warnings import warn

from prog_models import LinearModel

from . import state_estimator
from ..uncertain_data import MultivariateNormalDist, UncertainData

class KalmanFilter(state_estimator.StateEstimator):
    """
    A Kalman Filter (KF) for state estimation

    This class defines the logic for performing a kalman filter with a LinearModel (see Prognostics Model Package). This filter uses measurement data with noise to generate a state estimate and covariance matrix. 

    The supported configuration parameters (keyword arguments) for UKF construction are described below:

    Args:
        model (PrognosticsModel):
            A prognostics model to be used in state estimation
            See: Prognostics Model Package
        x0 (UncertainData, model.StateContainer, or dict):
            Initial (starting) state, with keys defined by model.states \n
            e.g., x = ScalarData({'abc': 332.1, 'def': 221.003}) given states = ['abc', 'def']

    Keyword Args:
        alpha (float, optional):
            KF Scaling parameter. An alpha > 1 turns this into a fading memory filter.
        t0 (float, optional):
            Starting time (s)
        dt (float, optional):
            time step (s)
        Q (list[list[float]], optional):
            Kalman Process Noise Matrix 
        R (list[list[float]], optional):
            Kalman Measurement Noise Matrix
    """
    default_parameters = {
        'alpha': 1, 
        't0': -1e-10,
        'dt': 1
    } 
    
    def __init__(self, model, x0, **kwargs):
        # Note: Measurement equation kept in constructor to keep it consistent with other state estimators. This way measurement equation can be provided as an ordered argument, and will just be ignored here
        if not isinstance(model, LinearModel):
            raise Exception('Kalman Filter only supports Linear Models (i.e., models derived from prog_models.LinearModel)')

        super().__init__(model, x0, **kwargs)

        self.x0 = x0

        if 'Q' not in self.parameters:
            self.parameters['Q'] = np.diag([1.0e-3 for i in x0.keys()])
        if 'R' not in self.parameters:
            # Size of what's being measured (not output) 
            # This is determined by running the measure function on the first state
            self.parameters['R'] = np.diag([1.0e-3 for i in range(model.n_outputs)])
        
        num_states = len(x0.keys())
        num_inputs = model.n_inputs + 1
        num_measurements = model.n_outputs
        F = deepcopy(model.A)
        B = deepcopy(model.B)
        if np.size(B) == 0:
            # If B is empty, replace with E. 
            # Append wont work if B is empty
            B = deepcopy(model.E)
        else:
            B = np.append(B, deepcopy(model.E), 0)

        self.filter = kalman.KalmanFilter(num_states, num_measurements, num_inputs)

        self.__state_keys = list(x0.keys())
        if isinstance(x0, dict) or isinstance(x0, model.StateContainer):
            warn("Warning: Use UncertainData type if estimating filtering with uncertain data.")
            self.filter.x = np.array([[x0[key]] for key in model.states]) # x0.keys()
            self.filter.P = self.parameters['Q'] / 10
        elif isinstance(x0, UncertainData):
            x_mean = x0.mean
            self.filter.x = np.array([[x_mean[key]] for key in model.states])

            # Reorder covariance to be in same order as model.states
            mapping = {i: list(x0.keys()).index(key) for i, key in enumerate(model.states)}
            cov = x0.cov  # Set covariance in case it has been calculated
            mapped_cov = [[cov[mapping[i]][mapping[j]] for j in range(len(cov))] for i in range(len(cov))] # Set covariance based on mapping
            self.filter.P = np.array(mapped_cov)
        else:
            raise TypeError("TypeError: x0 initial state must be of type {{dict, UncertainData}}")

        self.filter.Q = self.parameters['Q']
        self.filter.R = self.parameters['R']
        self.filter.F = F
        self.filter.B = B

    def estimate(self, t : float, u, z):
        """
        Perform one state estimation step (i.e., update the state estimate)

        Parameters
        ----------
        t : double
            Current timestamp in seconds (≥ 0.0)
            e.g., t = 3.4
        u : dict
            Measured inputs, with keys defined by model.inputs.
            e.g., u = {'i':3.2} given inputs = ['i']
        z : dict
            Measured outputs, with keys defined by model.outputs.
            e.g., z = {'t':12.4, 'v':3.3} given inputs = ['t', 'v']
        """
        assert t > self.t, "New time must be greater than previous"

        dt = t - self.t
        # Create u array, ensuring order of model.inputs. And reshaping to (n,1), n can be 0.
        inputs = np.array([u[key] for key in self.model.inputs]).reshape((-1,1))

        # Add row of ones (to account for constant E term)
        if np.size(inputs) == 0:
            inputs = np.array([[1]])
        else:
            inputs = np.append(inputs, [[1]], 0)

        self.t = t

        # Update equations
        # prog_models is dx = Ax + Bu + E
        # kalman_models is x' = Fx + Bu, where x' is the next state
        # Therefore we need to add the diagnol matrix 1 to A to convert
        # And A and B should be multiplied by the time step
        B = np.multiply(self.filter.B, dt) 
        F = np.multiply(self.filter.F, dt) + np.diag([1]* self.model.n_states)

        # Predict
        self.filter.predict(u = inputs, B = B, F = F)

        # Create z array, ensuring order of model.outputs
        outputs = np.array([z[key] for key in self.model.outputs])

        # Subtract D from outputs
        # This is done because prog_models expects the form: 
        #   z = Cx + D
        # While kalman expects
        #   z = Cx
        outputs = outputs - self.model.D

        self.filter.update(outputs, H=self.model.C)
    
    @property
    def x(self) -> MultivariateNormalDist:
        """
        Getter for property 'x', the current estimated state. 

        Example
        -------
        state = observer.x
        """
        return MultivariateNormalDist(self.model.states, self.filter.x.ravel(), self.filter.P, _type = self.model.StateContainer)



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from filterpy.monte_carlo import residual_resample
from numpy import array, empty, take, exp, max, take
from scipy.stats import norm
from warnings import warn

from prog_models.utils.containers import DictLikeMatrixWrapper

from . import state_estimator
from ..uncertain_data import UnweightedSamples, ScalarData, UncertainData
from ..exceptions import ProgAlgTypeError


class ParticleFilter(state_estimator.StateEstimator):
    """
    Estimates state using a Particle Filter (PF) algorithm.

    This class defines logic for a PF using a Prognostics Model (see Prognostics Model Package). This filter uses measurement data with noise to estimate the state of the system using a particles. At each step, particles are predicted forward (with noise). Particles are resampled with replacement from the existing particles according to how well the particles match the observed measurements.

    The supported configuration parameters (keyword arguments) for UKF construction are described below:

    Args:
        model (PrognosticsModel):
            A prognostics model to be used in state estimation
            See: Prognostics Model Package
        x0 (UncertainData, model.StateContainer, or dict):
            Initial (starting) state, with keys defined by model.states \n
            e.g., x = ScalarData({'abc': 332.1, 'def': 221.003}) given states = ['abc', 'def']

    Keyword Args:
        t0 (float, optional):
            Starting time (s)
        dt (float, optional): 
            time step (s)
        num_particles (int, optional):
            Number of particles in particle filter
        resample_fcn (function, optional):
            Resampling function ([weights]) -> [indexes] e.g., filterpy.monte_carlo.residual_resample
    """
    default_parameters = {
            't0': -1e-99,  # practically 0, but allowing for a 0 first estimate
            'num_particles': 20, 
            'resample_fcn': residual_resample,
        }

    def __init__(self, model, x0, **kwargs):
        super().__init__(model, x0, **kwargs)
        
        self._measure = model.output

        # Build array inplace
        if isinstance(x0, DictLikeMatrixWrapper) or isinstance(x0, dict):
            x0 = ScalarData(x0)
        elif not isinstance(x0, UncertainData):
            raise ProgAlgTypeError(f"ProgAlgTypeError: x0 must be of type UncertainData or StateContainer, was {type(x0)}.")

        sample_gen = x0.sample(self.parameters['num_particles'])
        samples = [array(sample_gen.key(k)) for k in x0.keys()]
        
        self.particles = model.StateContainer(array(samples))

        if 'R' in self.parameters:
            # For backwards compatibility
            warn("'R' is deprecated. Use 'measurement_noise' instead.", DeprecationWarning)
            self.parameters['measurement_noise'] = self.parameters['R']
        elif 'measurement_noise' not in self.parameters:
            self.parameters['measurement_noise'] = {key: 0.0 for key in x0.keys()}
    
    def __str__(self):
        return "{} State Estimator".format(self.__class__)
        
    def estimate(self, t : float, u, z):
        assert t > self.t, "New time must be greater than previous"
        dt = t - self.t
        self.t = t

        # Check Types
        if isinstance(u, dict):
            u = self.model.InputContainer(u)
        if isinstance(z, dict):
            z = self.model.OutputContainer(z)

        # Optimization
        particles = self.particles
        next_state = self.model.next_state
        apply_process_noise = self.model.apply_process_noise
        output = self._measure
        # apply_measurement_noise = self.model.apply_measurement_noise
        noise_params = self.model.parameters['measurement_noise']
        num_particles = self.parameters['num_particles']
        # Check which output keys are present (i.e., output of measurement function)
        measurement_keys = output(self.model.StateContainer({key: particles[key][0] for key in particles.keys()})).keys()
        zPredicted = {key: empty(num_particles) for key in measurement_keys}

        if self.model.is_vectorized:
            # Propagate particles state
            self.particles = apply_process_noise(next_state(particles, u, dt), dt)

            # Get particle measurements
            zPredicted = output(self.particles)
        else:
            # Propogate and calculate weights
            for i in range(num_particles):
                x = self.model.StateContainer({key: particles[key][i] for key in particles.keys()})
                x = next_state(x, u, dt) 
                x = apply_process_noise(x, dt)
                for key in particles.keys():
                    self.particles[key][i] = x[key]
                z = output(x)
                for key in measurement_keys:
                    zPredicted[key][i] = z[key]

        # Calculate pdf values
        pdfs = array([norm(zPredicted[key], noise_params[key]).logpdf(z[key])
                      for key in zPredicted.keys()])

        # Calculate log weights
        log_weights = pdfs.sum(0)

        # Scale
        # We subtract the max log weights for numerical stability. 
        # Sometimes log weights can be a large negative value
        # when you exponentiate that value the computer will round the result to 0 for most of the weights (sometimes all of them) 
        # this causes problems when trying to sample from the particles. 
        # We shift them up by the max log weight (essentially making the max log weight 0) to help avoid that problem. 
        # When we normalize the weights by dividing by the sum of all the weights, that constant cancels out.
        max_log_weight = max(log_weights)
        scaled_weights = log_weights - max_log_weight

        # Convert to weights
        unnorm_weights = exp(scaled_weights)
        
        # Normalize
        total_weight = sum(unnorm_weights)
        self.weights = unnorm_weights / total_weight

        # Resample indices
        indexes = self.parameters['resample_fcn'](self.weights)

        # Resampled particles
        samples = [take(self.particles[state], indexes)
                   for state in self.particles.keys()]

        # Particles as a dictionary
        self.particles = self.model.StateContainer(array(samples))

    @property
    def x(self) -> UnweightedSamples:
        """
        Getter for property 'x', the current estimated state. 

        Example
        -------
        state = observer.x
        """
        return UnweightedSamples(self.particles, _type = self.model.StateContainer)



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

"""
This file includes functions for calculating general metrics (i.e. mean, std, percentiles, etc.) on any distribution of type UncertainData (e.g. states, event_states, an EOL distribution, etc.)
"""
from typing import Iterable, Union
from numpy import isscalar, mean, std, array
from scipy import stats
from warnings import warn

from ..uncertain_data import UncertainData, UnweightedSamples

def calc_metrics(data : UncertainData, ground_truth : Union[float, dict] = None, **kwargs) -> dict:
    """Calculate all time of event metrics

    Args:
        data (array[float] or UncertainData): data from a single event
        ground_truth (float, optional): Ground truth value. Defaults to None. dict when data is  of type UncertainData.
        **kwargs (optional): Configuration parameters. Supported parameters include:
          * n_samples (int): Number of samples to use for calculating metrics (if data is not UnweightedSamples). Defaults to 10,000.
          * keys (list of strings, optional): Keys to calculate metrics for. Defaults to all keys.

    Returns:
        dict: collection of metrics
    """
    params = {
        'n_samples': 10000,  # Default is enough to get every percentile
    }
    params.update(kwargs)
    
    if isinstance(data, UncertainData):
        # Default to all keys
        keys = params.setdefault('keys', data.keys())
        if isinstance(keys, str):
            keys = [keys]
        
        if ground_truth and isscalar(ground_truth):
            # If ground truth is scalar, create dict (expected below)
            ground_truth = {key: ground_truth for key in keys}

        if isinstance(data, UnweightedSamples):
            samples = data
        else:
            # Some other distribution besides unweighted samples
            # Generate Samples
            samples = data.sample(params['n_samples'])

        # If unweighted_samples, calculate metrics for each key
        result = {key: calc_metrics(samples.key(key), 
                ground_truth if not ground_truth else ground_truth[key],  # If ground_truth is a dict, use key
                **kwargs) for key in keys}

        # Set values specific to distribution
        for key in keys:
            result[key]['mean'] = data.mean[key]
            result[key]['median'] = data.median[key]
            result[key]['percentiles']['50'] = data.median[key]

        return result
    elif isinstance(data, Iterable):
        if len(data) == 0:
            raise ValueError('Data must not be empty')
        # Is list or array
        if isscalar(data[0]) or data[0] is None:
            # list of numbers - this is the case that we can calculate
            pass
        elif isinstance(data[0], dict):
            # list of dicts - Supported for backwards compatabilities
            data = UnweightedSamples(data)
            return calc_metrics(data, ground_truth, **kwargs)
        else:
            raise TypeError("Data must be type Uncertain Data or array of dicts, was {}".format(type(data)))
    else:
        raise TypeError("Data must be type Uncertain Data or array of dicts, was {}".format(type(data)))

    # If we get here then Data is a list of numbers- calculate metrics for numbers
    data_abridged = array([d for d in data if d is not None]) # Must be array
    if len(data_abridged) == 0:
        raise ValueError('All samples were none')
    if len(data_abridged) < len(data):
        warn("Some samples were None, resulting metrics only consider non-None samples. Note: in some cases, this will bias the metrics.")
    data_abridged.sort()
    m = mean(data_abridged)
    median = data_abridged[int(len(data_abridged)/2)]
    metrics = {
        'min': data_abridged[0],
        'percentiles': {
            '0.01': data_abridged[int(len(data_abridged)/10000)] if len(data_abridged) >= 10000 else None,
            '0.1': data_abridged[int(len(data_abridged)/1000)] if len(data_abridged) >= 1000 else None,
            '1': data_abridged[int(len(data_abridged)/100)] if len(data_abridged) >= 100 else None,
            '10': data_abridged[int(len(data_abridged)/10)] if len(data_abridged) >= 10 else None,
            '25': data_abridged[int(len(data_abridged)/4)] if len(data_abridged) >= 4 else None,
            '50': median,
            '75': data_abridged[int(3*len(data_abridged)/4)] if len(data_abridged) >= 4 else None,
        },
        'median': median,
        'mean': m,
        'std': std(data_abridged),
        'max': data_abridged[-1],
        'median absolute deviation': sum([abs(x - median) for x in data_abridged])/len(data_abridged),
        'mean absolute deviation':   sum([abs(x - m)   for x in data_abridged])/len(data_abridged),
        'number of samples': len(data_abridged)
    }

    if ground_truth is not None:
        # Metrics comparing to ground truth
        metrics['mean absolute error'] = sum([abs(x - ground_truth) for x in data_abridged])/len(data_abridged)
        metrics['mean absolute percentage error'] = metrics['mean absolute error']/ ground_truth
        metrics['relative accuracy'] = 1 - abs(ground_truth - metrics['mean'])/ground_truth
        metrics['ground truth percentile'] = stats.percentileofscore(data_abridged, ground_truth)

    return metrics



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from numpy import array
from typing import List

def plot_hist(samples : array, fig : Figure = None, keys : List[str] = None, **kwargs) -> Figure:
    """Create a histogram

    Args:
        samples (Array(Dict)): A set of samples (e.g., states or eol estimates)
        fig (MatPlotLib Figure, optional): Existing histogram figure to be overritten. Defaults to create new figure.
        keys (List[String], optional): Keys to be plotted. Defaults to All.
    """

    # Input checks
    if len(samples) <= 0:
        raise Exception('Must include atleast one sample to plot')
    
    if keys is not None:
        if isinstance(keys, str):
            keys = [keys]
        try:
            iter(keys)
        except TypeError:
            raise TypeError("Keys should be a list of strings (e.g., ['state1', 'state2'], was {}".format(type(keys)))
        
        for key in keys:
            if key not in samples[0].keys():
                raise TypeError("Key {} was not present in samples (keys: {})".format(key, list(samples[0].keys())))
    else:
        keys = samples[0].keys()
    keys = list(keys)
    
    # Handle input
    parameters = {  # defaults
        'legend': True
    }
    parameters.update(kwargs)        

    if fig is None:
        # If no figure provided, create one
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        ax = fig.axes[0]
    
    # Plot
    for key in keys:
        ax.hist([sample[key] for sample in samples if sample[key] is not None], label=key, **kwargs)

    # Set legend
    if parameters['legend']:
        ax.legend().remove()  # Remove any existing legend - prevents "ghost effect"
        ax.legend(loc='upper right')
    
    return fig



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.
import matplotlib.pyplot as plt
from collections import UserDict
from typing import Dict
import numpy as np

from prog_algs.uncertain_data import UncertainData 

class ToEPredictionProfile(UserDict):
    """
    Data structure for storing the result of multiple predictions, including time of prediction. This data structure can be treated as a dictionary of time of prediction to Time of Event (ToE) prediction. Iteration of this data structure is in order of increasing time of prediction
    """
    def add_prediction(self, time_of_prediction: float, toe_prediction: UncertainData):
        """Add a single prediction to the profile

        Args:
            time_of_prediction (float): Time that the prediction was made
            toe_prediction (UncertainData): Distribution of predicted ToEs
        """
        self[time_of_prediction] = toe_prediction

    # Functions below are defined to ensure that any iteration is in order of increasing time of prediction
    def __iter__(self):
        return iter(sorted(super(ToEPredictionProfile, self).__iter__()))

    def items(self):
        """
        Get iterators for the items (time_of_prediction, toe_prediction) of the prediction profile
        """
        return iter((k, self[k]) for k in self)

    def keys(self):
        """
        Get iterator for the keys (i.e., time_of_prediction) of the prediction profile
        """
        return sorted(super(ToEPredictionProfile, self).keys())

    def values(self):
        """
        Get iterator for the values (i.e., toe_prediction) of the prediction profile
        """
        return [self[k] for k in self.keys()]

    def alpha_lambda(self, ground_truth : Dict[str, float], lambda_value : float, alpha : float, beta : float, **kwargs) -> Dict[str, bool]:
        """Calculate Alpha lambda metric for the prediction profile

        Args:
            ground_truth (dict[str, float]):
                Ground Truth time of event for each event (e.g., {'event1': 748, 'event2', 2233, ...})
            lambda_value (float):
                Prediction time at or after which metric is evaluated. Evaluation occurs at this time (if a prediction exists) or the next prediction following.
            alpha (float): 
                percentage bounds around time to event (where 0.2 allows 20% error TtE)
            beta (float):
                portion of prediction that must be within those bounds

        Keyword Args:
            keys (list[str], optional): 
                list of keys to use. If not provided, all keys are used.
            print (bool, optional)
                If True, print the results to the screen. Default is False.

        Returns:
            dict[str, bool]: If alpha lambda was met for each key (e.g., {'event1': True, 'event2', False, ...})
        """
        from ..metrics import alpha_lambda
        return alpha_lambda(self, ground_truth, lambda_value, alpha, beta, **kwargs)

    def prognostic_horizon(self, criteria_eqn, ground_truth, **kwargs) -> Dict[str, float]:
        """
        Compute prognostic horizon metric, defined as the difference between a time ti, when the predictions meet specified performance criteria, and the time corresponding to the true Time of Event (ToE), for each event.

        :math:`PH = ToE - ti`

        Args:
            toe_profile (ToEPredictionProfile): A profile of predictions, the combination of multiple predictions
            criteria_eqn (Callable function): A function (toe: UncertainData, ground_truth: dict[str, float]) -> dict[str, bool] calculating whether a prediction in ToEPredictionProfile meets some criteria. \n
                | Args: 
                |  * toe (UncertainData): A single prediction of Time of Event (ToE)
                |  * ground truth (dict[str, float]): Ground truth passed into prognostics_horizon
                | Returns: Map of event names to boolean representing if the event has been met. 
                |   e.g., {'event1': True, 'event2': False}
            ground_truth (dict): Dictionary containing ground truth; specified as key, value pairs for event and its value. E.g, {'event1': 47.3, 'event2': 52.1, 'event3': 46.1}
        
        Keyword Args:
            print (bool): 
                Boolean specifying whether the prognostic horizon metric should be printed.

        Returns:
            dict: Dictionary containing prognostic horizon calculations (value) for each event (key). e.g., {'event1': 12.3, 'event2': 15.1}
        """
        from ..metrics import prognostic_horizon
        return prognostic_horizon(self, criteria_eqn, ground_truth, **kwargs)

    def cumulative_relative_accuracy(self, ground_truth, **kwargs) -> Dict[str, float]:
        r"""
        Compute cumulative relative accuracy for a given profile, defined as the normalized sum of relative prediction accuracies at specific time instances.
        
        :math:`CRA = \Sigma \left( \dfrac{RA}{N} \right)` for each event

        Where :math:`\Sigma` is summation of all relative accuracies for a given profile and N is the total count of profiles [0]_

        Args:
            ground_truth (dict): Dictionary containing ground truth; specified as key, value pairs for event and its value. E.g, {'event1': 47.3, 'event2': 52.1, 'event3': 46.1}

        Returns:
            dict: Dictionary containing cumulative relative accuracy (value) for each event (key). e.g., {'event1': 12.3, 'event2': 15.1}

        References:
            .. [0] Journal Prognostics Health Management, Saxena et al.
        """
        from ..metrics import cumulative_relative_accuracy
        return cumulative_relative_accuracy(self, ground_truth, **kwargs)

    def monotonicity(self, **kwargs) -> Dict[str, float]:
        r"""Calculate monotonicty for a prediction profile. 
        Given a prediction profile, for each prediction: go through all predicted states and compare those to the next one.
        Calculates monotonicity for each prediction key using its associated mean value in :py:class:`prog_algs.uncertain_data.UncertainData`.
        
        :math:`monotonoicity = \|\Sigma \left( \dfrac{sign(i+1 - i)}{N-1} \right) \|`

        Where N is number of measurements and sign indicates sign of calculation. [0]_ [1]_

        Args:
            toe_profile (ToEPredictionProfile): A profile of predictions, the combination of multiple predictions
        
        Returns:
            dict (str, dict): Dictionary where keys represent a profile and dict is a subdictionary representing an event and its respective monotonicitiy value between [0, 1].

        References:
            .. [1] Coble, J., et. al. (2021). Identifying Optimal Prognostic Parameters from Data: A Genetic Algorithms Approach. Annual Conference of the PHM Society. http://www.papers.phmsociety.org/index.php/phmconf/article/view/1404
            .. [2] Baptistia, M., et. al. (2022). Relation between prognostics predictor evaluation metrics and local interpretability SHAP values. Aritifical Intelligence, Volume 306. https://www.sciencedirect.com/science/article/pii/S0004370222000078
        """
        from ..metrics import monotonicity
        return monotonicity(self, **kwargs)
    
    def plot(self, ground_truth : dict = None , alpha : float = None, show : bool = True) -> dict: # use ground truth, alpha if given,
        """Produce an alpha-beta plot depicting the TtE distribution by time of prediction for each event.

        Args:
            ground_truth (dict):
                Optional dictionary argument containing event and its respective ground truth value; none by default and plotted if specified
            alpha (float):
                Optional alpha value; none by default and plotted if specified
            show (bool):
                Optional bool value; specify whether to display generated plots. Default is true
        
        Returns:
            dict[str, Figure] :
                Collection of generated matplotlib figures for each event in profile\n
                e.g., {'event1': Fig, 'event2': Fig}
        
        Example:
            ::

                gt = {'event1': 3442, 'event2': 175} # Ground Truth
                figs = profile.plot(gt) # Figure with ground truth line
                figs = profile.plot(gt, alpha = 0.2) # Figure with ground truth line and 20% alpha bounds
                figs = profile.plot(gt, alpha = 0.2, show=False) # Dont display figure
        """
        result_figs = {}
        for t,v in self.items():
            raw_samples = v.sample(100) # sample distribution (red scatter plot)
            for key in v.keys():
                if key not in result_figs:
                    # Prepare Figure for Plot
                    fig_window = plt.figure() # Create new figure for this event key
                    fig_sub = fig_window.subplots()
                    fig_sub.grid()
                    fig_sub.set_title(f"{key} Event")
                    fig_sub.set_xlabel('Time of Prediction (s)') # time to prediction
                    fig_sub.set_ylabel('Time to Event (s)') # time to event
                    result_figs[key] = fig_window
                # Create scatter plot for this event
                samples = [e[key]-t for e in raw_samples]
                result_figs[key].get_axes()[0].scatter([t]*len(samples), samples, color='red') # Adding single distribution of estimates

        if ground_truth: # If ground_truth is specified, add ground_truth to each event plot (green line)
            for key, val in ground_truth.items():
                gt_x = range(int(val))
                gt_y = range(int(val), 0, -1)
                result_figs[key].get_axes()[0].plot(gt_x, gt_y, color='green')
                if alpha: # if ground_truth and alpha are specified, add alpha bounds (faded green highlight)
                    result_figs[key].get_axes()[0].fill_between(gt_x, np.array(gt_y)*(1-alpha), np.array(gt_y)*(1+alpha), color='green', alpha=0.2)
                result_figs[key].get_axes()[0].set_xlim(0, val+1)

        if show: # Optionally not display plots and just return plot objects
            plt.show()
        return result_figs 



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

__all__ = ['predictors', 'uncertain_data', 'state_estimators', 'run_prog_playback', 'metrics']
from . import predictors, state_estimators, uncertain_data

import warnings

__version__ = '1.4.2'

def run_prog_playback(obs, pred, future_loading, output_measurements, **kwargs):
    warnings.warn("Depreciated in 1.2.0, will be removed in a future release.", DeprecationWarning)
    config = {# Defaults
        'predict_rate': 0, # Default- predict every step
        'num_samples': 10,
        'predict_config': {}
    }
    config.update(kwargs)

    next_predict = output_measurements[0][0] + config['predict_rate']
    times = []
    inputs = []
    states = []
    outputs = []
    event_states = []
    toes = []
    index = 0
    for (t, measurement) in output_measurements:
        obs.estimate(t, future_loading(t), measurement)
        if t >= next_predict:
            pred_results = pred.predict(obs.x.sample(config['num_samples']), future_loading, **config['predict_config'])
            times.append(pred_results.times)
            inputs.append(pred_results.inputs)
            states.append(pred_results.states)
            outputs.append(pred_results.outputs)
            event_states.append(pred_results.event_states)
            toes.append(pred_results.time_of_event)
            index += 1
            next_predict += config['predict_rate']
    return predictors.PredictionResults(
        times, 
        inputs, 
        states, 
        outputs, 
        event_states, 
        toes
    ) 



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

from .uncertain_data import UncertainData
from .unweighted_samples import UnweightedSamples
from .scalar_data import ScalarData
from .multivariate_normal_dist import MultivariateNormalDist

__all__ = ['UncertainData', 'UnweightedSamples', 'ScalarData', 'MultivariateNormalDist']



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from filterpy import kalman
from numpy import diag, array
from warnings import warn

from . import state_estimator
from ..uncertain_data import MultivariateNormalDist, UncertainData

class UnscentedKalmanFilter(state_estimator.StateEstimator):
    """
    An Unscented Kalman Filter (UKF) for state estimation

    This class defines logic for performing an unscented kalman filter with a Prognostics Model (see Prognostics Model Package). This filter uses measurement data with noise to generate a state estimate and covariance matrix. 

    The supported configuration parameters (keyword arguments) for UKF construction are described below:

    Args:
        model (PrognosticsModel):
            A prognostics model to be used in state estimation
            See: Prognostics Model Package
        x0 (UncertainData, model.StateContainer, or dict):
            Initial (starting) state, with keys defined by model.states \n
            e.g., x = ScalarData({'abc': 332.1, 'def': 221.003}) given states = ['abc', 'def']

    Keyword Args:
        alpha (float, optional):
            UKF Scaling parameter
        beta (float, optional):
            UKF Scaling parameter
        kappa (float, optional):
            UKF Scaling parameter
        t0 (float, optional):
            Starting time (s)
        dt (float, optional):
            time step (s)
        Q (list[list[float]], optional):
            Process Noise Matrix 
        R (list[list[float]], optional):
            Measurement Noise Matrix 
    """
    default_parameters = {
        'alpha': 1, 
        'beta': 0, 
        'kappa': -1,
        't0': -1e-10,
        'dt': 1
    } 

    def __init__(self, model, x0, **kwargs):
        super().__init__(model, x0, **kwargs)

        self.__input = None
        self.x0 = x0
        # Saving for reduce pickling

        def measure(x):
            x = model.StateContainer({key: value for (key, value) in zip(x0.keys(), x)})
            R_err = model.parameters['measurement_noise'].copy()
            model.parameters['measurement_noise'] = dict.fromkeys(R_err, 0)
            z = model.output(x)
            model.parameters['measurement_noise'] = R_err
            return array(list(z.values())).ravel()

        if 'Q' not in self.parameters:
            self.parameters['Q'] = diag([1.0e-3 for i in x0.keys()])

        def state_transition(x, dt):
            x = model.StateContainer({key: value for (key, value) in zip(x0.keys(), x)})
            Q_err = model.parameters['process_noise'].copy()
            model.parameters['process_noise'] = dict.fromkeys(Q_err, 0)
            x = model.next_state(x, self.__input, dt)
            return array(list(x.values())).ravel()

        num_states = len(x0.keys())
        num_measurements = model.n_outputs
        points = kalman.MerweScaledSigmaPoints(num_states, alpha=self.parameters['alpha'], beta=self.parameters['beta'], kappa=self.parameters['kappa'])
        self.filter = kalman.UnscentedKalmanFilter(num_states, num_measurements, self.parameters['dt'], measure, state_transition, points)
        
        if isinstance(x0, dict) or isinstance(x0, model.StateContainer):
            warn("Use UncertainData type if estimating filtering with uncertain data.")
            self.filter.x = array(list(x0.values()))
            self.filter.P = self.parameters['Q'] / 10
        elif isinstance(x0, UncertainData):
            x_mean = x0.mean
            self.filter.x = array(list(x_mean.values()))
            self.filter.P = x0.cov
        else:
            raise TypeError("TypeError: x0 initial state must be of type {{dict, UncertainData}}")

        if 'R' not in self.parameters:
            # Size of what's being measured (not output) 
            # This is determined by running the measure function on the first state
            self.parameters['R'] = diag([1.0e-3 for i in range(len(measure(self.filter.x)))])
        self.filter.Q = self.parameters['Q']
        self.filter.R = self.parameters['R']

    def estimate(self, t : float, u, z):
        """
        Perform one state estimation step (i.e., update the state estimate)

        Parameters
        ----------
        t : double
            Current timestamp in seconds (≥ 0.0)
            e.g., t = 3.4
        u : dict
            Measured inputs, with keys defined by model.inputs.
            e.g., u = {'i':3.2} given inputs = ['i']
        z : dict
            Measured outputs, with keys defined by model.outputs.
            e.g., z = {'t':12.4, 'v':3.3} given inputs = ['t', 'v']
        """
        assert t > self.t, "New time must be greater than previous"
        dt = t - self.t
        self.__input = u
        self.t = t
        self.filter.predict(dt=dt)
        self.filter.update(array(list(z.values())))
    
    @property
    def x(self) -> MultivariateNormalDist:
        """
        Getter for property 'x', the current estimated state. 

        Example
        -------
        state = observer.x
        """
        return MultivariateNormalDist(self.x0.keys(), self.filter.x, self.filter.P, _type = self.model.StateContainer)



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name = 'prog_algs',
    version = '1.4.2.rc0',
    description = "The NASA Prognostics Algorithm Package is a framework for model-based prognostics (computation of remaining useful life) of engineering systems. It includes algorithms for state estimation and prediction, including uncertainty propagation. The algorithms use prognostic models (see prog_models) to perform estimation and prediction. The package enables rapid development of prognostics solutions for given models of components and systems. Algorithms can be swapped for comparative studies and evaluations",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url = 'https://nasa.github.io/progpy/prog_algs_guide.html',
    author = 'Christopher Teubert',
    author_email = 'christopher.a.teubert@nasa.gov',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Manufacturing', 
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: Other/Proprietary License ',   
        'Programming Language :: Python :: 3',     
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only'
    ],
    keywords = ['prognostics', 'diagnostics', 'fault detection', 'fdir', 'prognostics and health management', 'PHM', 'health management', 'ivhm'],
    package_dir = {"":"src"},
    packages = find_packages(where = 'src'),
    python_requires='>=3.7, <3.11',
    install_requires = [
        'numpy',
        'scipy',
        'filterpy',
        'matplotlib',
        'prog_models>=1.4.0'
    ],
    license = 'NOSA',
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/nasa/prog_algs/issues',
        'Docs': 'https://nasa.github.io/progpy/prog_algs_guide.html',
        'Organization': 'https://www.nasa.gov/content/diagnostics-prognostics',
        'Source': 'https://github.com/nasa/prog_algs',
    }
)



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from .plot_scatter import plot_scatter
from .plot_hist import plot_hist
__all__ = ['plot_scatter', 'plot_hist']



# Copyright © 2021 United States Government as represented by the Administrator of the
# National Aeronautics and Space Administration.  All Rights Reserved.

from collections import defaultdict
from typing import Union

MAX_COLUMN_WIDTH = 5 # numerical value will actually be min MAX_COLUMN_WIDTH-2 due to allocating spaces

def print_table_recursive(input_dict : dict, title : str, print_bool : bool = True) -> defaultdict:
    """
    Prints a table where keys are column headers and values are items in a row. 
    Returns the table formatted as a dictionary of tables represented by a list of str.

    Arguments
    ---------
    input_dict : dict
        A dictionary of keys and values to print out in a table. Values can be dictionaries.
    title : str
        Title of the table, printed before data rows.
    print_flag : bool = True
        An optional boolean value determining whether the generated table is printed.
    """
    row_list = _print_table_recursive_helper([], input_dict, title)[:-7]
    sub_tables = defaultdict(list)
    new_sub_table = []
    for row in row_list:
        new_sub_table.append(row)
        if len(new_sub_table) == 7:
            if sub_tables[len(new_sub_table[0])]:
                sub_tables[len(new_sub_table[0])].extend([new_sub_table[5], new_sub_table[6]])
            else:
                sub_tables[len(new_sub_table[0])].extend(new_sub_table)
            new_sub_table = []

    if print_bool:
        for k in sorted(sub_tables.keys(), reverse=True):
            print(*sub_tables[k], sep='\n')
    return sub_tables

def _set_width(max_width : int, input_value : Union[float, int]) -> str:
    if input_value < (10**max_width):
        ndigits = len(str(input_value))
        return f"{input_value:^{ndigits}.{max_width-ndigits}f}"
    else:
        scientific_input = f"{input_value:e}"
        split_e = scientific_input.split("e+")
        num_space = max_width - len(str(split_e[1])) - 2
        split_e[0] = str(split_e[0])[:num_space]
        return f"{split_e[0]}e+{split_e[1]}"
    # using this approach because e+ can't be formatted with f"{x:{some_width}g}"
    # what happens if we have 9.999999e+100 but are limited to 5?
    # the exponent itself will occupy 5, leaving no space for the numbers in front

def _print_table_recursive_helper(table_prog : list, input_dict : dict, title : str, key : str = None) -> list:
    """
    Helper function to recursively build subtables as a list of str.

    Arguments
    ---------
    table_prog : list
        A list of the table built so far. List of strings, where each string is a printable representation of a row.
    input_dict : dict
        A dictionary of keys and values to print out in a table. Values can be dictionaries.
    title : str
        Title of the table, printed before data rows.
    key : str = None
        Key for a value row, identifying what event the values belong to.
    """
    col_name_row = "| key |"
    value_row = f"| {str(key):^3} |"
    for k,v in input_dict.items():
        if isinstance(v, dict):
            if key != None:
                to_pass = key
                _print_table_recursive_helper(table_prog, v, f"{title} {k}", to_pass)
            else:
                to_pass = k
                _print_table_recursive_helper(table_prog, v, f"{title}", to_pass)
        else:
            col_len = len(max(str(k), str(v))) + 2
            col_name_row += f"{str(k):^{col_len}}|"
            if isinstance(v, (int, float)):
                adj_width = _set_width(MAX_COLUMN_WIDTH-2, v)
                value_row += f"{adj_width:^{col_len}}|"
            else:
                value_row += f"{str(v):^{col_len}}|"

    break_row = "+{}+".format((len(col_name_row)-2)*'-')
    title_row = f"+{title:^{len(break_row)-2}}+".title()
    table_prog.extend([break_row, title_row, break_row, col_name_row, break_row, value_row, break_row])
    
    return table_prog

